/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2018 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 5.02a                                            *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit.c for S12Z CPU and S12ZVMEVB eval board.
Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock frequency settings (configuration)
*/
#ifndef   OS_FSYS                     /* CPU Main clock frequency   */
  #define OS_FSYS        (25000000u)  /* Assume 25 MHz              */
#endif

#ifndef   OS_PCLK_TIMER               /* Peripheral clock for timer */
  #define OS_PCLK_TIMER  (OS_FSYS/2)  /* May vary from CPU clock    */
#endif                                /* depending on CPU           */

#ifndef   OS_PCLK_UART                /* Peripheral clock for UART  */
  #define OS_PCLK_UART   (OS_FSYS)    /* May vary from CPU clock    */
#endif                                /* depending on CPU           */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ   (1000u)
#endif

#define OS_TIMER_RELOAD  (OS_PCLK_TIMER / OS_TICK_FREQ)

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_DISABLED
#endif

/****** End of configurable options *********************************/

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/

//
// Gate Drive Unit
//
#define GDU_BASE_ADDR     (0x6A0u)
#define GDUF              (*(volatile OS_U8*) (GDU_BASE_ADDR + 0x06u))

//
// Clock, Reset and Power Management Unit
//
#define CPMU_BASE_ADDR    (0x6C0u)
#define RFLG              (*(volatile OS_U8*) (CPMU_BASE_ADDR + 0x03u))
#define SYNR              (*(volatile OS_U8*) (CPMU_BASE_ADDR + 0x04u))
#define REFDIV            (*(volatile OS_U8*) (CPMU_BASE_ADDR + 0x05u))
#define POSTDIV           (*(volatile OS_U8*) (CPMU_BASE_ADDR + 0x06u))
#define IFLG              (*(volatile OS_U8*) (CPMU_BASE_ADDR + 0x07u))
#define OSC               (*(volatile OS_U8*) (CPMU_BASE_ADDR + 0x1Au))

//
// Timer Module (TIM16B4CV3)
//
#define TIM_BASE_ADDR     (0x5C0u)
#define TIOS              (*(volatile OS_U8*) (TIM_BASE_ADDR + 0x00u))
#define TCNT              (*(volatile OS_U16*)(TIM_BASE_ADDR + 0x04u))
#define TSCR1             (*(volatile OS_U8*) (TIM_BASE_ADDR + 0x06u))
#define TTOV              (*(volatile OS_U8*) (TIM_BASE_ADDR + 0x07u))
#define TCTL2             (*(volatile OS_U8*) (TIM_BASE_ADDR + 0x09u))
#define TIE               (*(volatile OS_U8*) (TIM_BASE_ADDR + 0x0Cu))
#define TSCR2             (*(volatile OS_U8*) (TIM_BASE_ADDR + 0x0Du))
#define TFLG1             (*(volatile OS_U8*) (TIM_BASE_ADDR + 0x0Eu))
#define TC0               (*(volatile OS_U16*)(TIM_BASE_ADDR + 0x10u))
#define OCPD              (*(volatile OS_U8*) (TIM_BASE_ADDR + 0x2Cu))
#define PTPSR             (*(volatile OS_U8*) (TIM_BASE_ADDR + 0x2Eu))

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value
*
*  Return value
*    Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  OS_I16 Cnt;

  Cnt = TC0 - TCNT;
  if (Cnt > 0) {
  return (unsigned int)(OS_TIMER_RELOAD - Cnt);
  } else {
  return (unsigned int)0 - Cnt;
  }
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set
*
*  Return value
*    == 0; Interrupt pending flag not set
*    != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return ((unsigned int)(TFLG1 & 1u));
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       SysTick_Handler()
*
*  Function description
*    This is the code that gets called when the processor receives a
*    _SysTick exception. SysTick is used as OS timer tick.
*/
#ifdef __cplusplus
extern "C" {
#endif
interrupt void SysTick_Handler(void);
#ifdef __cplusplus
}
#endif
interrupt void SysTick_Handler(void) {
  TC0   += OS_TIMER_RELOAD;  // Set new compare value
  TFLG1 |= 1u;               // Reset interrupt request flag

  OS_INT_Enter();
  OS_TICK_Handle();
  OS_INT_Leave();
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialise the hardware (timer) required for embOS to run.
*    May be modified, if an other timer should be used.
*/
void OS_InitHW(void) {
  //
  // Structure with information about timer frequency, tick frequency, etc.
  // for micro second precise system time.
  //
  OS_SYSTIMER_CONFIG SysTimerConfig = {OS_PCLK_TIMER, OS_TICK_FREQ, 1, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
  OS_INT_IncDI();
  //
  // Ensure stable power supply before configuring the PLL
  //
  while ((GDUF & (1u << 5)) != 0) {
    GDUF |= (1u << 5);
  }
  //
  // Setup PLL: Generate 25 MHz core clock / 12.5 MHz bus clock from 4 MHz external oscillator
  //
  REFDIV  = ( 3u << 0)      // fREF = 4 MHz / (3 + 1) = 1 MHz
          | ( 0u << 6);     // 0 if fREF is from 0 MHz to 1 MHz
  SYNR    = (24u << 0)      // fVCO = fREF * 2 * (24 + 1) = 50 MHz
          | ( 1u << 6);     // 1 if fVCO is from 48 MHz to 80 MHz
  POSTDIV = ( 1u << 0);     // fPLL = fVCO (1 + 1) = 25 MHz
  OSC     = ( 1u << 7);     // Enable external oscillator
  while ((IFLG & (1u << 0)) == 0) {
    ;                       // Wait for external oscillator
  }
  while ((IFLG & (1u << 3)) == 0) {
    ;                       // Wait for PLL to lock
  }
  RFLG |= 0x60;             // Clear PORF and LVRF flags
  //
  // Initialise timer for embOS
  //
  TSCR1 = 0x08;             // Enable precision timer
  OCPD  = 0x0F;             // Disables the timer channel port
  TIOS  = 0x01;             // Timer Input Capture/Output Compare Select
  TCTL2 = 0x00;             // Timer Control Register 1
  TSCR2 = 0x00;             // Timer System Control Register 2
  TFLG1 = 0xFF;             // Clear flags
  PTPSR = 0x00;             // Prescale Factor = 1
  TC0   = OS_TIMER_RELOAD;  // Set Reload value
  TIE   = 0x01;             // Enable timer interrupt
  TSCR1 = 0x80;             // Enable timer
  //
  // Setup values for microsecond precise system time functions
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialise the optional communication for embOSView
  //
#if (OS_VIEW_IFSELECT != OS_VIEW_DISABLED)
  OS_COM_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This is basically the "core" of the idle loop.
*    This core loop can be changed, but:
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved. However, a simple program loop can be programmed
*    (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if (OS_DEBUG == 0)
      __asm(WAI);       // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*  Function description
*    This routine is required for task-info via embOSView or high
*    resolution time measurement functions.
*    It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_U32 Cnt;

  Time = OS_TIME_GetTicks32();
  Cnt  = ((OS_I16)OS_TIMER_RELOAD) - ((OS_I16)TC0 - (OS_I16)TCNT);
  //
  // Check if timer interrupt pending ...
  //
  if (TFLG1 & 1u) {
  Cnt  = ((OS_I16)OS_TIMER_RELOAD) - ((OS_I16)TC0 - (OS_I16)TCNT);
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*  Function description
*    Convert Cycles into microseconds.
*
*    If your clock frequency is not a multiple of 1 MHz,
*    you may have to modify this routine in order to get proper
*    diagnostics.
*
*    This routine is required for profiling or high resolution time
*    measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return (Cycles / (OS_TIMER_RELOAD / 1000u));
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Dummy routine.
*/
void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
}

#else
  #error "Selected embOSView interface is currently not supported."
#endif

/****** End Of File *************************************************/
